@@ -2,19 +2,26 @@
 # Unauthorized copying of this file, via any medium is strictly prohibited.
 import io
 import json
+import time
 import urllib.parse
 import zipfile
+import threading
 
 import requests
+from oauthlib.oauth2 import BackendApplicationClient, TokenExpiredError
+from requests_oauthlib import OAuth2Session
 
 from . import utils
 from .dataset import Dataset
-from .exceptions import PipeAlreadyExists, SystemAlreadyExists, ConfigUploadFailed
+from .exceptions import PipeAlreadyExists, SystemAlreadyExists, ConfigUploadFailed, \
+    TimeoutWhileWaitingForBackupToFinishException, ServiceUnavailable, InternalServerError
 from .log import Log
 from .pipe import Pipe
 from .system import System
-
+from .model import Model
 from .utils import validate_response_is_ok
+from requests.exceptions import RequestException
+
 
 def quote_id_for_url(item_id):
     return urllib.parse.quote(item_id, safe="")
@@ -28,35 +35,101 @@ class Connection:
                  username=None, password=None,
                  client_certificate=None,
                  timeout=30,
-                 jwt_auth_token=None, verify_ssl=True):
+                 retries=0,
+                 retry_delay=5,
+                 jwt_auth_token=None,
+                 client_id=None, client_secret=None, token_endpoint=None,
+                 verify_ssl=True):
+        """
+        client_id, client_secret and token_endpoint are specified when we want to use the OpenID connect
+        "client credentials" grant-type.
+        """
+        self._close_has_been_called_event = threading.Event()
+
         if not sesamapi_base_url.endswith("/"):
             sesamapi_base_url += "/"
+
         self.sesamapi_base_url = sesamapi_base_url
         self.username = username
         self.password = password
+        self._client_certificate = client_certificate
         self.timeout = timeout
+        self._jwt_auth_token = jwt_auth_token
+        self._verify_ssl = verify_ssl
+        self._retries = retries
+        self._retry_delay = retry_delay
 
-
-        if self.username:
-            auth = (self.username, self.password)
-        else:
-            auth = None
-
-        headers = {
+        self._headers = {
             "ACCEPT": "application/json,*/*"
         }
 
-        if jwt_auth_token:
-            headers["AUTHORIZATION"] = "Bearer %s" % (jwt_auth_token,)
+        self._oauth2_config = None
+        if client_id:
+            self._oauth2_config = {
+                "client_id": client_id,
+                "client_secret": client_secret,
+                "token_endpoint": token_endpoint,
+            }
 
-        self.session = session = requests.Session()
-        self.session.verify = verify_ssl
+        self._oauth2_client = None
+        self._oauth2_session = None
+        self._oauth2_token = None
+        self._session = None
 
-        if auth is not None:
-            session.auth = auth
-        if client_certificate is not None:
-            session.cert = client_certificate
-        session.headers = headers
+    def set_jwt_auth_token(self, jwt_auth_token):
+        """Replaces the current jwt_auth_token (if any) with the new jwt."""
+        self._jwt_auth_token = jwt_auth_token
+        # We must close any existing session, since the old session uses the old jwt token. A new session will
+        # be lazily created by the session() method when needed.
+        session = self._session
+        self._session = None
+        if session is not None:
+            session.close()
+
+    @property
+    def session(self):
+        if self._oauth2_config is not None:
+
+            oauth2_config = self._oauth2_config
+
+            if self._oauth2_client is None:
+                self._oauth2_client = BackendApplicationClient(client_id=oauth2_config["client_id"])
+
+            if self._oauth2_session is None:
+                self._oauth2_session = OAuth2Session(client=self._oauth2_client)
+                if self._headers:
+                    self._oauth2_session.headers = self._headers
+
+            # If no token has been created yet or if the previous token has expired (or is close to it),
+            # fetch a new access token before returning the session to the callee.
+            if self._oauth2_token is None or self._oauth2_token["expires_at"] <= (time.time() + 60):
+                self._oauth2_token = self._oauth2_session.fetch_token(
+                    token_url=oauth2_config["token_endpoint"],
+                    client_id=oauth2_config["client_id"],
+                    client_secret=oauth2_config["client_secret"],
+                )
+
+            return self._oauth2_session
+        else:
+            if self._session is None:
+                if self._jwt_auth_token:
+                    self._headers["AUTHORIZATION"] = "Bearer %s" % (self._jwt_auth_token,)
+
+                self._session = session = requests.Session()
+                session.verify = self._verify_ssl
+
+                if self.username:
+                    auth = (self.username, self.password)
+                else:
+                    auth = None
+
+                if auth is not None:
+                    session.auth = auth
+                if self._client_certificate is not None:
+                    session.cert = self._client_certificate
+                session.headers = self._headers
+
+            return self._session
 
     def __enter__(self):
         return self
@@ -65,7 +138,18 @@ class Connection:
         self.close()
 
     def close(self):
-        self.session.close()
+        # IS-8849: interrupt any pending do_* retries
+        self._close_has_been_called_event.set()
+
+        if self._session is not None:
+            self._session.close()
+        self._session = None
+
+        if self._oauth2_session is not None:
+            self._oauth2_session.close()
+        self._oauth2_session = None
+        self._oauth2_token = None
+        self._oauth2_client = None
 
     def _get_requests_kwargs(self):
         kwargs = {}
@@ -73,31 +157,170 @@ class Connection:
             kwargs["timeout"] = self.timeout
         return kwargs
 
-    def do_put_request(self, url, data=None, **kwargs):
+    def do_put_request(self, url, data=None, allowable_response_status_codes=(200,), **kwargs):
         assert url is not None
         session_put_kwargs = self._get_requests_kwargs()
         session_put_kwargs.update(kwargs)
-        return self.session.put(url, data=data, **session_put_kwargs)
+
+        retries = self._retries
+        while True:
+            try:
+                try:
+                    response = self.session.put(url, data=data, **session_put_kwargs)
+                except TokenExpiredError:
+                    # Manually clear the token and retry once. The token is supposed to be automatically
+                    # refreshed, but if the local clock is off that can mess things up.
+                    self._oauth2_token = None
+                    response = self.session.put(url, data=data, **session_put_kwargs)
+
+                if allowable_response_status_codes:
+                    utils.validate_response_is_ok(response, allowable_response_status_codes)
+
+                return response
+            except (RequestException, InternalServerError, ServiceUnavailable, ConnectionError) as e:
+                if retries <= 0:
+                    raise e
+
+                retries -= 1
+                self.interruptable_sleep(self._retry_delay)
 
     def do_get_request(self, url, allowable_response_status_codes=(200, 404), **kwargs):
         assert url is not None
         session_get_kwargs = self._get_requests_kwargs()
         session_get_kwargs.update(kwargs)
-        response = self.session.get(url, **session_get_kwargs)
-        utils.validate_response_is_ok(response, allowable_response_status_codes)
-        return response
 
-    def do_post_request(self, url, **kwargs):
+        retries = self._retries
+        while True:
+            try:
+                try:
+                    response = self.session.get(url, **session_get_kwargs)
+                except TokenExpiredError:
+                    # Manually clear the token and retry once. The token is supposed to be automatically
+                    # refreshed, but if the local clock is off that can mess things up.
+                    self._oauth2_token = None
+                    response = self.session.get(url, **session_get_kwargs)
+
+                if allowable_response_status_codes:
+                    utils.validate_response_is_ok(response, allowable_response_status_codes)
+
+                return response
+            except (RequestException, InternalServerError, ServiceUnavailable, ConnectionError) as e:
+                if retries <= 0:
+                    raise e
+
+                retries -= 1
+                self.interruptable_sleep(self._retry_delay)
+
+    def do_head_request(self, url, allowable_response_status_codes=(200, 404), **kwargs):
+        assert url is not None
+        session_get_kwargs = self._get_requests_kwargs()
+        session_get_kwargs.update(kwargs)
+
+        retries = self._retries
+        while True:
+            try:
+                try:
+                    response = self.session.head(url, **session_get_kwargs)
+                except TokenExpiredError:
+                    # Manually clear the token and retry once. The token is supposed to be automatically
+                    # refreshed, but if the local clock is off that can mess things up.
+                    self._oauth2_token = None
+                    response = self.session.get(url, **session_get_kwargs)
+
+                if allowable_response_status_codes:
+                    utils.validate_response_is_ok(response, allowable_response_status_codes)
+
+                return response
+            except (RequestException, InternalServerError, ServiceUnavailable, ConnectionError) as e:
+                if retries <= 0:
+                    raise e
+
+                retries -= 1
+                self.interruptable_sleep(self._retry_delay)
+
+    def do_options_request(self, url, allowable_response_status_codes=(200, 404), **kwargs):
+        assert url is not None
+        session_get_kwargs = self._get_requests_kwargs()
+        session_get_kwargs.update(kwargs)
+
+        retries = self._retries
+        while True:
+            try:
+                try:
+                    response = self.session.options(url, **session_get_kwargs)
+                except TokenExpiredError:
+                    # Manually clear the token and retry once. The token is supposed to be automatically
+                    # refreshed, but if the local clock is off that can mess things up.
+                    self._oauth2_token = None
+                    response = self.session.get(url, **session_get_kwargs)
+
+                if allowable_response_status_codes:
+                    utils.validate_response_is_ok(response, allowable_response_status_codes)
+
+                return response
+            except (RequestException, InternalServerError, ServiceUnavailable, ConnectionError) as e:
+                if retries <= 0:
+                    raise e
+
+                retries -= 1
+                self.interruptable_sleep(self._retry_delay)
+
+    def do_post_request(self, url, allowable_response_status_codes=(200,), **kwargs):
         assert url is not None
         session_post_kwargs = self._get_requests_kwargs()
         session_post_kwargs.update(kwargs)
-        return self.session.post(url, **session_post_kwargs)
 
-    def do_delete_request(self, url, **kwargs):
+        retries = self._retries
+        while True:
+            try:
+                try:
+                    response = self.session.post(url, **session_post_kwargs)
+                except TokenExpiredError:
+                    # Manually clear the token and retry once. The token is supposed to be automatically
+                    # refreshed, but if the local clock is off that can mess things up.
+                    self._oauth2_token = None
+                    response = self.session.post(url, **session_post_kwargs)
+
+                if allowable_response_status_codes:
+                    utils.validate_response_is_ok(response, allowable_response_status_codes)
+
+                return response
+            except (RequestException, InternalServerError, ServiceUnavailable, ConnectionError) as e:
+                if retries <= 0:
+                    raise e
+
+                retries -= 1
+                self.interruptable_sleep(self._retry_delay)
+
+    def do_delete_request(self, url, allowable_response_status_codes=(200,), **kwargs):
         assert url is not None
         session_delete_kwargs = self._get_requests_kwargs()
         session_delete_kwargs.update(kwargs)
-        return self.session.delete(url, **session_delete_kwargs)
+
+        retries = self._retries
+        while True:
+            try:
+                try:
+                    response = self.session.delete(url, **session_delete_kwargs)
+                except TokenExpiredError:
+                    # Manually clear the token and retry once. The token is supposed to be automatically
+                    # refreshed, but if the local clock is off that can mess things up.
+                    self._oauth2_token = None
+                    response = self.session.delete(url, **session_delete_kwargs)
+
+                if allowable_response_status_codes:
+                    utils.validate_response_is_ok(response, allowable_response_status_codes)
+
+                return response
+            except (RequestException, InternalServerError, ServiceUnavailable, ConnectionError) as e:
+                if retries <= 0:
+                    raise e
+
+                retries -= 1
+                self.interruptable_sleep(self._retry_delay)
+
+    def interruptable_sleep(self, sleep_time):
+        self._close_has_been_called_event.wait(sleep_time)
 
     @property
     def pipes_url(self):
@@ -142,6 +365,9 @@ class Connection:
     def get_system_config_url(self, system_id):
         return self.get_system_url(system_id) + "/config"
 
+    def get_system_status_url(self, system_id):
+        return self.get_system_url(system_id) + "/status"
+
     def get_system_metadata_url(self, system_id):
         return self.get_system_url(system_id) + "/metadata"
 
@@ -155,6 +381,13 @@ class Connection:
     def systemconfigs_url(self):
         return self.sesamapi_base_url + "system-prototypes"
 
+    @property
+    def models_url(self):
+        return self.sesamapi_base_url + "models"
+
+    def get_model_url(self, model_id):
+        return self.models_url + "/" + quote_id_for_url(model_id)
+
     @property
     def stats_url(self):
         return self.sesamapi_base_url + "stats"
@@ -185,6 +418,10 @@ class Connection:
     def node_metadata_url(self):
         return self.sesamapi_base_url + "metadata"
 
+    @property
+    def node_namespaces_url(self):
+        return self.sesamapi_base_url + "namespaces"
+
     def get_secret_url(self, key, system_id=None):
         if system_id is None:
             return self.secrets_url + "/" + quote_id_for_url(key)
@@ -202,9 +439,11 @@ class Connection:
     def datasets_url(self):
         return self.sesamapi_base_url + "datasets"
 
-    @property
-    def config_url(self):
-        return self.sesamapi_base_url + "config"
+    def get_config_url(self, config_group=None):
+        url = self.sesamapi_base_url + "config"
+        if config_group is not None:
+            url += "/" + urllib.parse.quote(config_group, safe="")
+        return url
 
     def get_dataset_url(self, dataset_id):
         return self.datasets_url + "/" + quote_id_for_url(dataset_id)
@@ -228,7 +467,7 @@ class Connection:
         request_params = {}
         if include_internal_systems:
             request_params["include-internal-systems"] = "true"
-        response = self.do_get_request(self.systems_url, allowable_response_status_codes=[200])
+        response = self.do_get_request(self.systems_url, allowable_response_status_codes=[200], params=request_params)
         parsed_response = utils.parse_json_response(response)
 
         for system_json in parsed_response:
@@ -243,6 +482,24 @@ class Connection:
         system_json = utils.parse_json_response(response)
         return System(self, system_json)
 
+    def get_models(self):
+        models = []
+
+        response = self.do_get_request(self.models_url, allowable_response_status_codes=[200])
+        parsed_response = utils.parse_json_response(response)
+
+        for model_json in parsed_response:
+            models.append(Model(self, model_json))
+
+        return models
+
+    def get_model(self, model_id):
+        response = self.do_get_request(self.get_model_url(model_id))
+        if response.status_code == 404:
+            return None
+        model_json = utils.parse_json_response(response)
+        return Model(self, model_json)
+
     def get_systemconfigs(self):
         response = self.do_get_request(self.systemconfigs_url, allowable_response_status_codes=[200])
         return utils.parse_json_response(response)
@@ -312,8 +569,9 @@ class Connection:
 
     def add_pipes(self, pipe_configs, force=False):
         request_params = {"force": "true" if force else "false"}
-        response = self.do_post_request(self.pipes_url, json=pipe_configs, params=request_params)
-        utils.validate_response_is_ok(response, allowable_response_status_codes=[201, 409, 400])
+        response = self.do_post_request(self.pipes_url, allowable_response_status_codes=[201, 409, 400],
+                                        json=pipe_configs, params=request_params)
+
         if response.status_code == 409:
             raise PipeAlreadyExists(response.text)
 
@@ -327,8 +585,9 @@ class Connection:
 
     def add_systems(self, system_configs, force=False):
         request_params = {"force": "true" if force else "false"}
-        response = self.do_post_request(self.systems_url, json=system_configs, params=request_params)
-        utils.validate_response_is_ok(response, allowable_response_status_codes=[201, 409, 400])
+        response = self.do_post_request(self.systems_url, allowable_response_status_codes=[201, 409, 400],
+                                        json=system_configs, params=request_params)
+
         if response.status_code == 409:
             raise SystemAlreadyExists(response.text)
 
@@ -340,6 +599,18 @@ class Connection:
         systems = [System(self, system_json) for system_json in parsed_response]
         return systems
 
+    def add_models(self, model_configs):
+        response = self.do_post_request(self.models_url, allowable_response_status_codes=[201, 409, 400],
+                                        json=model_configs)
+
+        parsed_response = utils.parse_json_response(response)
+
+        if response.status_code == 400:
+            raise ConfigUploadFailed(response=response)
+
+        models = [Model(self, model_json) for model_json in parsed_response]
+        return models
+
     def get_logs(self):
         logs = []
 
@@ -362,13 +633,16 @@ class Connection:
     def get_metadata(self):
         """Gets the current metadata for the node"""
         response = self.do_get_request(self.node_metadata_url)
-        metadata = utils.parse_json_response(response)
-        return metadata
+        return utils.parse_json_response(response)
 
     def set_metadata(self, metadata):
         """Replaces the metadata for the node with the specified dictionary"""
-        response = self.do_put_request(self.node_metadata_url, json=metadata)
-        utils.validate_response_is_ok(response, allowable_response_status_codes=[200])
+        self.do_put_request(self.node_metadata_url, allowable_response_status_codes=[200], json=metadata)
+
+    def get_namespaces(self):
+        """Gets the current metadata for the node"""
+        response = self.do_get_request(self.node_namespaces_url, allowable_response_status_codes=[200])
+        return utils.parse_json_response(response)
 
     def get_secrets(self, include_internal_secrets=False):
         """This returns a list with the names of the secrets."""
@@ -376,8 +650,7 @@ class Connection:
         if include_internal_secrets:
             request_params["include-internal-secrets"] = "true"
 
-        response = self.do_get_request(self.secrets_url, params=request_params)
-        utils.validate_response_is_ok(response, allowable_response_status_codes=[200])
+        response = self.do_get_request(self.secrets_url, allowable_response_status_codes=[200], params=request_params)
         secrets_info = utils.parse_json_response(response)
         return secrets_info
 
@@ -385,56 +658,51 @@ class Connection:
         """This returns a dict with secretname=>secretvalue mappings. This is only available for nodes that
         are running in ci-test mode. In addition, the user has to have the "SystemAdmin" role."""
         request_params = {}
-        response = self.do_get_request(self.secrets_values_url, params=request_params)
-        utils.validate_response_is_ok(response, allowable_response_status_codes=[200])
+        response = self.do_get_request(self.secrets_values_url, allowable_response_status_codes=[200],
+                                       params=request_params)
         secrets_info = utils.parse_json_response(response)
         return secrets_info
 
     def put_secrets(self, secrets_content, dont_encrypt=False):
         encrypt = dont_encrypt and "false" or "true"
-        response = self.do_put_request(self.secrets_url, json=secrets_content, params=dict(encrypt=encrypt))
-        utils.validate_response_is_ok(response, allowable_response_status_codes=[200])
+        response = self.do_put_request(self.secrets_url, allowable_response_status_codes=[200],
+                                       json=secrets_content, params=dict(encrypt=encrypt))
 
         secrets_info = utils.parse_json_response(response)
         return secrets_info
 
     def post_secrets(self, secrets_content, dont_encrypt=False):
         encrypt = dont_encrypt and "false" or "true"
-        response = self.do_post_request(self.secrets_url, json=secrets_content, params=dict(encrypt=encrypt))
-        utils.validate_response_is_ok(response, allowable_response_status_codes=[200])
+        response = self.do_post_request(self.secrets_url, allowable_response_status_codes=[200],
+                                        json=secrets_content, params=dict(encrypt=encrypt))
 
         secrets_info = utils.parse_json_response(response)
         return secrets_info
 
     def delete_secret(self, key):
-        response = self.do_delete_request(self.get_secret_url(key))
-        utils.validate_response_is_ok(response, allowable_response_status_codes=[200])
+        response = self.do_delete_request(self.get_secret_url(key), allowable_response_status_codes=[200])
         secrets_info = utils.parse_json_response(response)
         return secrets_info
 
     def get_env_vars(self):
-        response = self.do_get_request(self.env_vars_url)
-        utils.validate_response_is_ok(response, allowable_response_status_codes=[200])
+        response = self.do_get_request(self.env_vars_url, allowable_response_status_codes=[200])
         env_vars_info = utils.parse_json_response(response)
         return env_vars_info
 
     def put_env_vars(self, env_vars_content):
-        response = self.do_put_request(self.env_vars_url, json=env_vars_content)
-        utils.validate_response_is_ok(response, allowable_response_status_codes=[200])
+        response = self.do_put_request(self.env_vars_url, allowable_response_status_codes=[200], json=env_vars_content)
 
         env_vars_info = utils.parse_json_response(response)
         return env_vars_info
 
     def post_env_vars(self, env_vars_content):
-        response = self.do_post_request(self.env_vars_url, json=env_vars_content)
-        utils.validate_response_is_ok(response, allowable_response_status_codes=[200])
+        response = self.do_post_request(self.env_vars_url, allowable_response_status_codes=[200], json=env_vars_content)
 
         env_vars_info = utils.parse_json_response(response)
         return env_vars_info
 
     def delete_env_var(self, key):
-        response = self.do_delete_request(self.get_env_var_url(key))
-        utils.validate_response_is_ok(response, allowable_response_status_codes=[200])
+        response = self.do_delete_request(self.get_env_var_url(key), allowable_response_status_codes=[200])
         env_vars_info = utils.parse_json_response(response)
         return env_vars_info
 
@@ -446,40 +714,45 @@ class Connection:
 
     def set_license(self, license_key_content):
         """Replaces the license for the node with the specified license key"""
-        response = self.do_put_request(self.node_license_url, data=license_key_content)
-        utils.validate_response_is_ok(response, allowable_response_status_codes=[200])
+        response = self.do_put_request(self.node_license_url, allowable_response_status_codes=[200],
+                                       data=license_key_content)
         license_info = utils.parse_json_response(response)
         return license_info
 
     def delete_metadata(self):
         """Deleted all metadata for the node"""
-        response = self.do_delete_request(self.node_metadata_url)
-        utils.validate_response_is_ok(response, allowable_response_status_codes=[200])
+        self.do_delete_request(self.node_metadata_url, allowable_response_status_codes=[200])
 
-    def get_config(self):
+    def get_config(self, config_group=None):
         """Returns the configuration of the sesam node.
+        :param config_group: The config-group to get the config from. The default is None, meaning the main config
+                      of the node.
         """
-        response = self.do_get_request(self.config_url, allowable_response_status_codes=[200])
+        response = self.do_get_request(self.get_config_url(config_group), allowable_response_status_codes=[200])
         utils.validate_equal_case_insensitive(response.headers["Content-Type"], "application/json")
         config = utils.parse_json_response(response)
         return config
 
-    def get_config_as_json_string(self):
+    def get_config_as_json_string(self, config_group=None):
         """Returns the configuration of the sesam node as a json-string, using the same formatting
         as the raw server response.
+        :param config_group: The config-group to get the config from. The default is None, meaning the main config
+                      of the node.
         """
-        response = self.do_get_request(self.config_url, allowable_response_status_codes=[200])
+        response = self.do_get_request(self.get_config_url(config_group), allowable_response_status_codes=[200])
         utils.validate_equal_case_insensitive(response.headers["Content-Type"], "application/json")
-        return response.text
+        return response.content.decode("utf-8")
 
-    def get_config_as_zip(self):
+    def get_config_as_zip(self, config_group=None):
         """Returns the configuration of the sesam node as a zip-archive with separate files for each pipe, system,
         etc.
+        :param config_group: The config-group to get the config from. The default is None, meaning the main config
+                      of the node.
 
         Tip: the returned bytes can be parsed like this:
                config = zipfile.ZipFile(io.BytesIO(returned_bytes)
         """
-        response = self.do_get_request(self.config_url,
+        response = self.do_get_request(self.get_config_url(config_group),
                                        allowable_response_status_codes=[200],
                                        headers={"ACCEPT": "application/zip"},
                                        )
@@ -487,6 +760,7 @@ class Connection:
         return response.content
 
     def upload_config(self, config,
+                      config_group=None,
                       force=False,
                       _use_multipart_form_data=True):
         """
@@ -497,10 +771,12 @@ class Connection:
         :param config: The config. This can be either a filename, a file-like object, or a bytes-object. The format
                       must be either a json object (as returned by the get_config()-method, or a zip-archive (as
                       returned by the get_config_as_zip()-method).
+        :param config_group: The config-group to upload the config to. The default is None, meaning the main config
+                      of the node.
         :param force: If this is False (the default), no changes will be made if the configuration contains any errors.
                       If it is True, the node will attempt to apply the new configuration on a best-effort basis, and
                       component with erronous configuration will get their "is-valid-config" flag set to false.
-        :param _post_as_multipart_form_data: This can be set to False in order to post the config content as the
+        :param _use_multipart_form_data: This can be set to False in order to post the config content as the
                                 request-body, instead of posting it using the normal "multipart/form-data" encoding.
                                 This is only used in ci-tests.
 
@@ -554,8 +830,10 @@ class Connection:
             if _use_multipart_form_data:
                 # This is the normal case. We use the same method a browser would use when uploading a file.
                 files = {'file': ('config', theconfigfile, content_type)}
-                response = self.do_put_request(self.config_url,
+                response = self.do_put_request(self.get_config_url(config_group),
                                                params=request_params,
+                                               # we will deal with any 400-errors later
+                                               allowable_response_status_codes=[200, 400],
                                                files=files)
             else:
                 # We want to put the config content directly in the response body.
@@ -565,18 +843,16 @@ class Connection:
                         raise TypeError(
                             "The content-type is 'application/zip', but the data isn't of type 'bytes': type(data):%s"
                                 % (type(data),))
-                response = self.do_put_request(self.config_url, data=data,
+                response = self.do_put_request(self.get_config_url(config_group), data=data,
                                                params=request_params,
+                                               # we will deal with any 400-errors later
+                                               allowable_response_status_codes=[200, 400],
                                                headers={"Content-Type": content_type})
 
         finally:
             if theconfigfile_must_be_closed:
                 theconfigfile.close()
 
-        utils.validate_response_is_ok(response,
-                                      allowable_response_status_codes=[200,
-                                                                       400  # we will deal with any 400-errors later
-                                                                       ])
         utils.validate_equal_case_insensitive(response.headers["Content-Type"], "application/json")
 
         parsed_response = utils.parse_json_response(response)
@@ -605,16 +881,53 @@ class Connection:
         the get_access_control_list() method for examples.
         """
         acl_url = self.sesamapi_base_url + "permissions/" + acl_path
-        response = self.do_put_request(acl_url, json=acl)
-        utils.validate_response_is_ok(response, allowable_response_status_codes=[200])
+        self.do_put_request(acl_url, allowable_response_status_codes=[200], json=acl)
 
     def login(self, email, password):
         login_url = self.sesamapi_base_url + "jwt"
 
-        response = self.do_post_request(login_url, data={"email": email, "password": password})
-
-        validate_response_is_ok(response, [200])
+        response = self.do_post_request(login_url, allowable_response_status_codes=[200],
+                                        data={"email": email, "password": password})
 
         jwt = response.text
 
         return jwt
+
+    def get_backup(self, backup_id):
+        url = self.sesamapi_base_url + "backups/" + backup_id
+        response = self.do_get_request(url, allowable_response_status_codes=[200])
+        backup_info = response.json()
+        return backup_info
+
+    def get_backups(self):
+        url = self.sesamapi_base_url + "backups"
+        response = self.do_get_request(url, allowable_response_status_codes=[200])
+        backups = response.json()
+        return backups
+
+    def trigger_backup(self):
+        url = self.sesamapi_base_url + "backups"
+        response = self.do_post_request(url, allowable_response_status_codes=[200])
+        backup_info = response.json()
+        return backup_info
+
+    def wait_for_backup_to_finish(self, backup_id, timeout=60, poll_interval=0.5):
+        """Utility method for waiting until the specified backup is no longer running.
+
+        :param timeout: The maximum time to wait for the backup to stop running. If the timeout is exceeded,
+                        an error is raised.
+        """
+        starttime = time.monotonic()
+        while True:
+            url = self.sesamapi_base_url + "backups/" + backup_id
+            response = self.do_get_request(url, allowable_response_status_codes=[200])
+            backup_info = response.json()
+            if backup_info["status"] != "running":
+                break
+            elapsedtime = time.monotonic() - starttime
+            if elapsedtime > timeout:
+                raise TimeoutWhileWaitingForBackupToFinishException(
+                    "Timed out while waiting for the backup to finish running! lat backup_info:\n%s" % (
+                        backup_info,))
+
+            time.sleep(poll_interval)
