@@ -26,8 +26,7 @@ class Pipe(EntityBase):
 
     def get_pump(self):
         url = self._connection.get_pipe_pump_url(self.id)
-        response = self._connection.do_get_request(url)
-        utils.validate_response_is_ok(response, [200])
+        response = self._connection.do_get_request(url, allowable_response_status_codes=[200])
         pump_json = response.json()
         return Pump(self, self._connection, pump_json)
 
@@ -78,12 +77,19 @@ class Pipe(EntityBase):
             params["limit"] = limit
 
         # The entities are streamed from the server, so we have to enable streaming here.
-        response = self._connection.do_get_request(url, stream=True, params=params)
-        utils.validate_response_is_ok(response)
+        response = self._connection.do_get_request(url, allowable_response_status_codes=[200],
+                                                   stream=True, params=params)
 
         return entity_json.parse_json_stream(response.raw)
 
-    def post_entities(self, entities):
+    def post_entities(self, entities,
+                      is_full=None,
+                      sequence_id=None,
+                      request_id=None,
+                      previous_request_id=None,
+                      is_last=None,
+                      is_first=None,
+                      ):
         """This posts the specified entities to the pipe. This only works for pipes that has a source
         of the type 'http_endpoint'.
         """
@@ -92,57 +98,56 @@ class Pipe(EntityBase):
         # If the size of the entities is very large, we have to use chunked encoding; otherwise we will get
         # a "connection reset" error.
         entities_as_json = entity_json.entities_to_json(entities)
+
         class BytesIOWithNoLength(io.BytesIO):
             def __len__(self):
                 raise io.UnsupportedOperation()
+
             def __bool__(self):
                 return True
+
         entities_stream = BytesIOWithNoLength(entities_as_json.encode("utf-8"))
 
+        params = {}
+        if is_full is not None:
+            params["is_full"] = True
+        if sequence_id is not None:
+            params["sequence_id"] = sequence_id
+        if request_id is not None:
+            params["request_id"] = request_id
+        if previous_request_id is not None:
+            params["previous_request_id"] = previous_request_id
+        if is_last is not None:
+            params["is_last"] = is_last
+        if is_first is not None:
+            params["is_first"] = is_first
+
         response = self._connection.do_post_request(url,
+                                                    allowable_response_status_codes=[200],
                                                     headers={"Content-Type": "application/json"},
+                                                    params=params,
                                                     data=entities_stream)
-        utils.validate_response_is_ok(response)
+
         return response.json()
 
     def delete(self):
         """Deletes this pipe from the sesam-node.
         """
         url = self._connection.get_pipe_url(self.id)
-        response = self._connection.do_delete_request(url)
-        utils.validate_response_is_ok(response, allowable_response_status_codes=[200])
-        self.update_raw_jsondata(response.text)
+        response = self._connection.do_delete_request(url, allowable_response_status_codes=[200])
+
+        self.update_raw_jsondata(response.content.decode("utf-8"))
 
     def modify(self, pipe_config, force=False):
         """Modifies the pipe with the specified configuration."""
         request_params = {"force": "true" if force else "false"}
         pipe_config = copy.deepcopy(pipe_config)
-        pipe_config["_id"] = self.id
-        response = self._connection.do_put_request(self._connection.get_pipe_config_url(self.id), json=pipe_config,
+        response = self._connection.do_put_request(self._connection.get_pipe_config_url(self.id),
+                                                   allowable_response_status_codes=[200, 400],
+                                                   json=pipe_config,
                                                    params=request_params)
-        utils.validate_response_is_ok(response, allowable_response_status_codes=[200, 400])
 
         if response.status_code == 400:
             raise ConfigUploadFailed(response=response)
 
-        self.update_raw_jsondata(response.text)
-
-    def get_metadata(self):
-        """Gets the current metadata for this pipe"""
-        url = self._connection.get_pipe_metadata_url(self.id)
-        response = self._connection.do_get_request(url)
-        utils.validate_response_is_ok(response, allowable_response_status_codes=[200])
-        metadata = response.json()
-        return metadata
-    
-    def set_metadata(self, metadata):
-        """Replaces the metadata for this pipe with the specified dictionary"""
-        url = self._connection.get_pipe_metadata_url(self.id)
-        response = self._connection.do_put_request(url, json=metadata)
-        utils.validate_response_is_ok(response, allowable_response_status_codes=[200])
-
-    def delete_metadata(self):
-        """Deleted all metadata for this pipe"""
-        url = self._connection.get_pipe_metadata_url(self.id)
-        response = self._connection.do_delete_request(url)
-        utils.validate_response_is_ok(response, allowable_response_status_codes=[200])
+        self.update_raw_jsondata(response.content.decode("utf-8"))
